/* The color indicator on per-item parameters */

.helipad_per_item .widget-label::before {
	content: '';
	height: 10px; width: 10px;
	border-radius: 10px;
	border: 1px solid rgba(0,0,0,0.2);
	display: inline-block;
	margin-right:5px;
}
.helipad_cpanel .widget-inline-hbox { max-width: calc(100% - 4px); }
.helipad_cpanel .widget-inline-hbox .widget-label { width: auto !important;} /* Not sure why the previous rule screws this up */

.helipad_cpanel div.widget-logslider { max-width: calc(100% - 50px); }

/* Turn an HBox of checkboxes into a wrapping Checkgrid */

.helipad_checkgrid .widget-label-basic input { display: none; }
.helipad_checkgrid .widget-label-basic span {
	display: block;
	padding: 3px 15px;
	cursor: pointer;
	width: 120px;
}
.helipad_checkgrid .widget-label-basic input:checked + span {
	background: #419BF9;
	color: #FFF;
}
.helipad_checkgrid .widget-label-basic input:enabled + span:hover { background: #DBDBDB; }
.helipad_checkgrid .widget-label-basic input:checked:enabled + span:hover { background: #1c87f8 }
.helipad_checkgrid .widget-label { display: none; }
.helipad_checkgrid .widget-checkbox {
	width: auto; height: auto;
	margin: 0 2px 2px 0;
}
.helipad_checkgrid .lm-Panel, .helipad_paramgroup .lm-Panel { overflow: visible; }
.helipad_checkgrid .widget-vbox, .helipad_paramgroup .widget-vbox {
	display: block;
	float: left;
}
.helipad_param_peritem .widget-inline-hbox { padding-right: 15px; }
.helipad_checkgrid .widget-label-basic input:disabled + span {
	opacity: 0.5;
	cursor: default;
}

.helipad_checkgrid .p-Panel { /* Position toggle-all correctly */
	overflow: visible;
	position: static;
}

.helipad_toggleAll {
	position: absolute;
	right: 0; top: -3.5em;
	background: none;
	color: #999;
	width: auto;
	border-radius: 5px;
	cursor: pointer;
}
.helipad_toggleAll:hover {
	background: #EEE;
	color: #555;
	box-shadow: none !important;
}

/* Progress bar */

.widget-hprogress { width: 100%; }
.indeterminate .progress-bar {
	background-image: linear-gradient(45deg,rgba(255,255,255,.15) 25%,transparent 25%,transparent 50%,rgba(255,255,255,.15) 50%,rgba(255,255,255,.15) 75%,transparent 75%,transparent);
	background-size: 40px 40px;
}
.helipad_running.indeterminate .progress-bar { animation: helipad-progress 2s linear infinite; }

@keyframes helipad-progress {
  from	{ background-position: 40px 0; }
  to	{ background-position: 0 0; }
}
.helipad_progress_area button { width: 90px; }

/* Misc */

.helipad_checkentry_func input[type="text"] { font-style: italic; }
.helipad_info {
	font-style: italic;
	border: 1px solid #CB2;
	background: #FFC;
	display: inline-block;
	padding: 5px 15px;
	border-radius: 4px;
	height: auto !important;
}
.helipad_info::before {
	content: '\26A0\FE0F ';
	font-style: normal;
	margin-right: 8px;
}
.helipad_info .widget-html-content { display: inline; }
.helipad_info code {
	background-color: rgba(0,0,0,0.075);
	padding: 1px 5px;
	font-size: 95%;
}

.helipad_cpanel.invalid > * {
	opacity: 0.25;
	pointer-events: none;
}
.helipad_cpanel .helipad_modal {
	opacity: 1;
	position: absolute;
	width: 80%; height: auto;
	text-align: center;
	top: 50px; left: 50%;
	margin: 0 0 0 -40%;
	padding: 20px;
	background: rgba(255,210,200,0.75);
	border: 1px solid rgba(100,0,0,0.5);
	border-radius: 4px;
	font-size: 18px;
	color: #422;
	pointer-events: all;
}
.helipad_modal::before {
	content: '\1F6D1 ';
	font-size: larger;
	margin-right: 12px;
}

.helipad_cpanel > .p-Widget { margin-bottom: 0.5em; }
.helipad_accordion_hidden { display: none; } /* For some reason Ipywidgets throws 'Javascript Error: t is undefined' when accessing accordion layouts */