"""
Helipad is an agent-based modeling framework with a hook-based architecture and powerful visualization capabilities. Import and instantiate the `Helipad` class to set up and launch a model. See https://helipad.dev for API documentation.
"""

from helipad.model import Helipad, MultiLevel
from helipad.agent import Agent, baseAgent
from helipad import utility

__version__ = "1.6.2"